@@ -29,6 +29,15 @@ import pyghmi.ipmi.sdr as sdr
 import socket
 import struct
 
+try:
+    range = xrange
+except NameError:
+    pass
+try:
+    buffer
+except NameError:
+    buffer = memoryview
+
 
 boot_devices = {
     'net': 4,
@@ -173,7 +182,7 @@ class Command(object):
             'device_id': response['data'][0],
             'device_revision': response['data'][1] & 0b1111,
             'manufacturer_id': struct.unpack(
-                '<I', struct.pack('3B', *response['data'][6:9]) + '\x00')[0],
+                '<I', struct.pack('3B', *response['data'][6:9]) + b'\x00')[0],
             'product_id': struct.unpack(
                 '<H', struct.pack('2B', *response['data'][9:11]))[0],
         }
@@ -433,6 +442,12 @@ class Command(object):
         :param duration: Set if wanting to request turn on for a duration
                          rather than indefinitely on
         """
+        self.oem_init()
+        try:
+            self._oem.set_identify(on, duration)
+            return
+        except exc.UnsupportedFunctionality:
+            pass
         if duration is not None:
             duration = int(duration)
             if duration > 255:
@@ -803,7 +818,7 @@ class Command(object):
         that
         """
         if self._netchannel is None:
-            for channel in chain((0xe, ), xrange(1, 0xc)):
+            for channel in chain((0xe, ), range(1, 0xc)):
                 try:
                     rsp = self.xraw_command(
                         netfn=6, command=0x42, data=(channel,))
@@ -927,7 +942,7 @@ class Command(object):
         numpol = ord(rsp['data'][1])
         desiredchandest = (channel << 4) | destination
         availpolnum = None
-        for polnum in xrange(1, numpol + 1):
+        for polnum in range(1, numpol + 1):
             currpol = self.xraw_command(netfn=4, command=0x13,
                                         data=(9, polnum, 0))
             polidx, chandest = struct.unpack_from('>BB', currpol['data'][2:4])
@@ -1002,7 +1017,7 @@ class Command(object):
                 parsedip = socket.inet_pton(socket.AF_INET, ip)
                 destdata.extend((0, 0))
                 destdata.extend(parsedip)
-                destdata.extend('\x00\x00\x00\x00\x00\x00')
+                destdata.extend(b'\x00\x00\x00\x00\x00\x00')
             except socket.error:
                 if self._supports_standard_ipv6:
                     parsedip = socket.inet_pton(socket.AF_INET6, ip)
@@ -1019,7 +1034,7 @@ class Command(object):
                 acknowledge_timeout is not None):
             currtype = self.xraw_command(netfn=0xc, command=2, data=(
                 channel, 18, destination, 0))
-            if currtype['data'][0] != '\x11':
+            if currtype['data'][0] != b'\x11':
                 raise exc.PyghmiException("Unknown parameter format")
             currtype = bytearray(currtype['data'][1:])
             if acknowledge_required is not None:
@@ -1048,7 +1063,7 @@ class Command(object):
         """Set the management controller identifier, per DCMI specification
 
         """
-        return self._chunkwise_dcmi_set(0xa, mci + '\x00')
+        return self._chunkwise_dcmi_set(0xa, mci + b'\x00')
 
     def get_asset_tag(self):
         """Get the system asset tag, per DCMI specification
@@ -1080,7 +1095,7 @@ class Command(object):
         return retstr
 
     def _chunkwise_dcmi_set(self, command, data):
-        chunks = [data[i:i+15] for i in xrange(0, len(data), 15)]
+        chunks = [data[i:i+15] for i in range(0, len(data), 15)]
         offset = 0
         for chunk in chunks:
             chunk = bytearray(chunk, 'utf-8')
@@ -1522,7 +1537,7 @@ class Command(object):
         if password:
             password = str(password)
             if 21 > len(password) > 16:
-                password = password.ljust(20, '\x00')
+                password = password.ljust(20, b'\x00')
                 data[0] |= 0b10000000
             elif len(password) > 20:
                 raise Exception('password has limit of 20 chars')
@@ -1669,7 +1684,7 @@ class Command(object):
             # however another convention that exists is all '\xff'
             # if this fails, pass up the error so that calling code knows
             # that the deletion did not go as planned for now
-            self.set_user_name(uid, '\xff' * 16)
+            self.set_user_name(uid, b'\xff' * 16)
         return True
 
     def disable_user(self, uid, mode):
