@@ -44,6 +44,14 @@ import pyghmi.util.webclient as wc
 import socket
 import struct
 import weakref
+try:
+    range = xrange
+except NameError:
+    pass
+try:
+    buffer
+except NameError:
+    buffer = memoryview
 
 inventory.register_inventory_category(cpu)
 inventory.register_inventory_category(dimm)
@@ -120,7 +128,11 @@ mac_format = '{0:02x}:{1:02x}:{2:02x}:{3:02x}:{4:02x}:{5:02x}'
 
 
 def _megarac_abbrev_image(name):
-    if len(name) <= 16:
+    # MegaRAC platform in some places needs an abbreviated filename
+    # Their scheme in such a scenario is a max of 20.  Truncation is
+    # acheived by taking the first sixteen, then skipping ahead to the last
+    # 4 (presumably to try to keep '.iso' or '.img' in the name).
+    if len(name) <= 20:
         return name
     return name[:16] + name[-4:]
 
@@ -271,6 +283,12 @@ class OEMHandler(generic.OEMHandler):
         return (self.oemid['manufacturer_id'], self.oemid['device_id'],
                 self.oemid['product_id']) in fpc_ids
 
+    @property
+    def is_sd350(self):
+        return (19046, 32, 13616) == (self.oemid['manufacturer_id'],
+                                      self.oemid['device_id'],
+                                      self.oemid['product_id'])
+
     @property
     def has_tsm(self):
         """True if this particular server have a TSM based service processor
@@ -336,7 +354,7 @@ class OEMHandler(generic.OEMHandler):
                 tmp_command = dict(catspec["command"])
                 tmp_command["data"] = list(tmp_command["data"])
                 count = 0
-                for i in xrange(0x01, 0xff):
+                for i in range(0x01, 0xff):
                     tmp_command["data"][-1] = i
                     try:
                         partrsp = self.ipmicmd.xraw_command(**tmp_command)
@@ -367,7 +385,7 @@ class OEMHandler(generic.OEMHandler):
                 )
             except Exception:
                 # If we can't parse an inventory category, ignore it
-                print traceback.print_exc()
+                print(traceback.print_exc())
                 continue
 
             for item in items:
@@ -377,7 +395,7 @@ class OEMHandler(generic.OEMHandler):
                     self.oem_inventory_info[key] = item
                 except Exception:
                     # If we can't parse an inventory item, ignore it
-                    print traceback.print_exc()
+                    print(traceback.print_exc())
                     continue
 
     def get_leds(self):
@@ -392,6 +410,12 @@ class OEMHandler(generic.OEMHandler):
                                         led_status_default)
                 yield (name, {'status': status})
 
+    def set_identify(self, on, duration):
+        if on and not duration and self.is_sd350:
+            self.ipmicmd.xraw_command(netfn=0x3a, command=6, data=(1, 1))
+        else:
+            raise pygexc.UnsupportedFunctionality()
+
     def process_fru(self, fru):
         if fru is None:
             return fru
@@ -469,7 +493,8 @@ class OEMHandler(generic.OEMHandler):
         if len(bdata['data'][:]) != 3:
             self._hasimm = False
             return False
-        self._hasimm = ord(bdata['data'][1]) & 1 == 1
+        rdata = bytearray(bdata['data'][:])
+        self._hasimm = (rdata[1] & 1 == 1) or (rdata[1] & 16 == 16)
         return self._hasimm
 
     def get_oem_firmware(self, bmcver):
@@ -614,7 +639,7 @@ class OEMHandler(generic.OEMHandler):
             else:
                 # fall back to a dumber, but more universal formatter
                 ipv6str = binascii.b2a_hex(ipv6_addr)
-                ipv6str = ':'.join([ipv6str[x:x+4] for x in xrange(0, 32, 4)])
+                ipv6str = ':'.join([ipv6str[x:x+4] for x in range(0, 32, 4)])
             netdata['ipv6_addresses'] = [
                 '{0}/{1}'.format(ipv6str, ipv6_prefix)]
 
@@ -672,7 +697,7 @@ class OEMHandler(generic.OEMHandler):
         self.ipmicmd.xraw_command(netfn=0x32, command=0x9f,
                                   data=(1, selector, 0, 1))
         # now do the set
-        for x in xrange(0, 256, 64):
+        for x in range(0, 256, 64):
             currdata = padded[x:x+64]
             currchunk = x // 64 + 1
             cmddata = [1, selector, currchunk] + currdata
@@ -686,7 +711,7 @@ class OEMHandler(generic.OEMHandler):
                                         data=(7, 1, 0))
         imgnames = rsp['data'][1:]
         shortnames = []
-        for idx in xrange(0, len(imgnames), 22):
+        for idx in range(0, len(imgnames), 22):
             shortnames.append(imgnames[idx+2:idx+22].rstrip('\0'))
         return shortnames
 
@@ -737,7 +762,10 @@ class OEMHandler(generic.OEMHandler):
                                   data=(1, 1))
 
     def attach_remote_media(self, url, username, password):
-        if self.has_megarac:
+        if self.has_imm:
+            imm.attach_remote_media(self.ipmicmd, self._certverify, url,
+                                    username, password)
+        elif self.has_megarac:
             proto, host, path = util.urlsplit(url)
             if proto == 'smb':
                 proto = 'cifs'
@@ -759,7 +787,9 @@ class OEMHandler(generic.OEMHandler):
                     raise
 
     def detach_remote_media(self):
-        if self.has_megarac:
+        if self.has_imm:
+            imm.detach_remote_media(self.ipmicmd, self._certverify)
+        elif self.has_megarac:
             self.ipmicmd.xraw_command(
                 netfn=0x32, command=0x9f, data=(8, 10, 0, 0))
             self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=(8, 11))
